// ======================================================================
// IR receiver, power switch, and LCD controller with USB interface.
// ======================================================================

// ======================================================================
// IR receiver
// 
// Read IR mark/space patterns from a TSOP1738 on the ICP input pin.
// Mark and space periods are measured in units of 12Mhz/1024, which is
// 85.33 us. Up to 35 periods are stored in a buffer, where they can
// be read via a USB request. The buffer layout and USB requests are
// compatible with the "IgorPlug-USB" device, which is supported by LIRC
// through the "igorplugusb" driver. This implementation should also
// work with that driver. If IgorPlug-USB compatibility is not required,
// the resolution could be improved by reducing SCALING to 9, and you
// could also get rid of the additional endpoint (USBTINY_ENDPOINT).
//
// There are some IR protocols with very long packets length, such as
// the "NEC" protocol (http://www.sbprojects.com/knowledge/ir/nec.htm).
// This protocol uses 67 mark/space periods, which is too much for the
// limited amount of RAM in the ATtiny2313. Simply truncating the
// packet to the amount of available buffer space is not an option,
// because the significant bits are at the end of the packet. Instead,
// a buffer of 35 bytes (IR_MAX) is used, of which the last 16 bytes
// (IR_WRAP) are used as a circular buffer. Because both the address
// (16 periods) and the command code (16 periods) are sent twice (normal
// and inverted), the wrapping will result in a buffer with both the
// device address and the inverted command code. Other IR protocols are
// typically shorter than 35 periods.
//
// 16-bit Timer1 is initialized in normal mode with a clock of 12MHz/8.
// Both the positive and negative edges of the IR input signal on the
// ICP input pin trigger an input capture interrupt, because the edge
// selection is toggled on every interrupt. Output compare register A
// is used for a timeout to detect the end of an IR transmission.
// The timeout value is updated on every edge. When at least 4 edges
// were detected when a timeout occurs, the packet count ir.count is
// incremented and the ir.length byte is set to the packed length. This
// byte is checked when a read request is received via USB to see if a
// packet is available
//
// The two interrupt handlers in this file run for more than 28 cycles,
// so they need to reenable interrupts to allow USB interrupts to be
// handled in time. To avoid nested IR interrupts, the input capture and
// output compare interrupts are disabled before the global interrupt
// flag is set. In the input capture interrupt handler, the interrupts
// are disabled for a maximum of 26, 27 and 21 cycles, and in the output
// compare interrupt handler for 21, 23 and 21 cycles. Unfortunately,
// it was not possible to meet the 28 cycle maximum for the input
// capture interrupt handler without replacing the compiler generated
// prolog and epilog code by inline assembly, which makes the code a
// little fragile.
//
// The IR receiver is disabled during an IR buffer read, so that a
// packet that is being tranmitted to the host, is not overwritten by
// the next packet. When a transfer is in progress, 'inpos' has a value
// other than 0xff. When the buffer is read completely, the IR receiver
// is reenabled again, but when the next IR packet has already started,
// this packed is discarded by setting 'edges' to 0xff.
//
// Stack space calculations for gcc-3.4.3 and gcc-4.1.0:
//
// 3.4.3   4.1.0
// -----   -----
//   0       0       main
//   0       0           lcd_init (inlined)
//   2       2           usb_init
//   0       0           ir_init (inlined)
//   6       5           usb_poll
//   0       0               usb_receive (inlined)
//   2       2                   usb_setup
//   5       5                   usb_out
//   4       4                       lcd_write
//   2       2                           lcd_read4
//   2       2                           lcd_write4
//   0       0               usb_transmit (inlined)
//   2       4                   usb_in
//   2       2                   crc
//
// Maximum stack space is 6+5+4+2=17 for gcc-3.4.3 and 5+5+4+2=16 for
// gcc-4.1.0. Total stack usage:
// - USB interrupt:			11 bytes
// - IR input capture interrupt:	11 bytes
// - IR output compare interrupt: 	 7 bytes
// - main():				17 bytes
// The two IR interrupt handler cannot be active simultaneously, so the
// maximum stack space is 17+11+11=39 bytes.
//
// NOTE: The above cycle and stack size figures were obtained by
//       inspecting the code generated by gcc-3.4.3 and gcc-4.1.0.
//       Other compiler versions may generate different code, leading
//       to different figures. Likewise, the replacement prolog and
//       epilog code may have to be adapted when a different compiler
//       version allocates different registers.
//
// Copyright 2006-2008 Dick Streefland
//
// This is free software, licensed under the terms of the GNU General
// Public License as published by the Free Software Foundation.
// ======================================================================

// ======================================================================
// LCD display
//
// There are two USB request codes to control the LCD, one to write
// instructions, and one to write data to the display. OUT transfers
// are used to transfer up to 255 bytes for a single request.
//
// LCD init is only performed completely, if LCD_INIT_PRESENT is defined.
// Making this switchable was necessary to get IR, LCD, SWITCH, and SWITCH_PROG
// fit into the tiny 2k program space.
// If LCD_INIT_PRESENT is NOT set, the display stays in 8bit mode, thus the
// byte sequence 0x32, 0x28, 0x01, 0x06, 0x0c has to be sent as the first(!)
// LCD_INSTR command.  0x20 instead of 0x28 for one lined displays.
//
// Failing to do so will probably render the LCD display unusable until the
// next AVR reset. Re-enabling should be possible (untested, though) by 0x33
// plus the init sequence, because this way the display should be put reliably
// into 8bit bus mode before switching back to 4bit. However, after the first
// nibble a 4.1ms delay is suggested, which is not done in the general byte
// write routine.
// The effect probably depends on the LCD chip vendor.
//
// Copyright 2006-2008 Dick Streefland
// ======================================================================

// ======================================================================
// IR Switch
// 
// Compares pulse/pause lengths against intervals in EEPROM, and sets a defined
// data pin high as long as the represented key is pressed (minimum: 250ms).
//
//
// EEMEM  0-34:   intervals_hi[IR_MAX]
//                upper interval limit (hi) for pulse/pauses 0-34
// EEMEM 35-69:   intervals_lo[IR_MAX]
//                lower interval limit (lo) for pulse/pauses 0-34
//
// End of packet (EoP) is indicated by hi=255,lo=0
// If the received signal is shorter than the programmed one or EoP is
// detected, the IR signal is considered accepted. The first condition is a
// side effect of the algorithm, and should probably be changed someday.
//
// The interval array is split this way for easier access.
// EEPROM access had to be hardcoded as a macro to reduce register use in the
// interrupt handler.
// Still, it was not possible to meet the 25 cycle maximum for the timer
// compare interrupt handler without replacing the compiler generated prolog
// and epilog code by inline assembly, which makes the code a little fragile.
//
//
// EEPROM contents (as defined above) can be programmed with the
// SWITCH_PROG usb command, if SWITCH_PROG_PRESENT is set.
// SWITCH_PROG only programs a single cell. Experiments with multiple
// locations showed that after approx. 7 cells weird things happened,
// probably because of usb_out() time constraints (which are unknown to me).
//
// inpos is (mis)used to indicate that usb_out() receives intervals instead of
// LCD commands.  Thus after (multiple) SWITCH_PROG a IGORPLUG_CLEAR must be
// issued to continue sampling (this might impose a race disabling the IR
// switch if the machine is switched off in between...).
//
//
// The function uses Timer1 (which is used for IR sampling) as well for
// disabling the switch after 250ms, thus clock had to be reduced to 12MHz/64.
// This affects the work of the noise canceler - which still has just enough
// time to work with SCALING=9.
// Actually, the routine is now much less sensitive to noise...
//
//
// Maximum interrupt disabled cycle lengths (gcc-4.3.3)
// > first number including last interrupted command, PC stack push, rjmp
// > all numbers including sei, and command after sei/reti (worst case: rts))
//
// - input capture handler (updated):	22, 25, 22 cycles
// - compare interrupt handler A:	16, 24, 22 cycles
// - compare interrupt handler B:	19         cycles
// (gee, a single instruction interrupt takes up to 19 cycles...)
//
//
// Stack space change (gcc-4.3.3):
// - usb_out:  w/  LCD_INIT_PRESENT     5 bytes
//   usb_out:  w/o LCD_INIT_PRESENT     2 bytes (due to inlined lcd_write)
// - lcd_write w/  LCD_INIT_PRESENT     2 bytes
//   lcd_write w/o LCD_INIT_PRESENT     0 bytes (inlined)
// - lcd_read4                          0 bytes (inlined)
// - lcd_write4                         0 bytes (inlined)
// - usb_in                             2 bytes
// Maximum non-interrupt stack space is 5+5+2=12 (LCD_INIT_PRESENT)
// or 5+2=7 (no LCD_INIT_PRESENT)
//
// - IR output compare interrupt:      11 bytes
// Maximum stack space is unchanged 12+11+11=34 bytes.
//
//
// Total amount of RAM used: 4 data + 85 bss + 34 stack = 123
//
// Copyright 2010 Matthias Hopf
// ======================================================================


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include "usb.h"

// ----------------------------------------------------------------------
// Feature support
// ----------------------------------------------------------------------

#define LCD_PRESENT	    1	// LCD support
#define LCD_INIT_PRESENT    0	// LCD is initialized completely

#define SWITCH_PRESENT	    1	// IR SWITCH support
#define SWITCH_PROG_PRESENT 1	// USB command to program SWITCH intervals


// ----------------------------------------------------------------------
// General
// ----------------------------------------------------------------------

#define DO_NAKED	1	// ONLY SET TO 0 DURING ASM CODE ANALYSIS
#define CLK_KHZ		12000
#define CLK_DIV_SHIFT	6	// 3: 1/8  6: 1/64

#if	DO_NAKED
#  define NAKED		naked
#  define NAKED_CODE(x) x
#else
#  define NAKED
#  define NAKED_CODE(x) (void)0
#endif
#define _MERGE(x,y)	_MERGE2(x,y)
#define _MERGE2(x,y)    x ## y
#define US_2_CLK(x)	((_MERGE(CLK_KHZ,LL) * (x)) / 1000 >> CLK_DIV_SHIFT)

#if	SWITCH_PRESENT
// output compare 1A + 1B match interrupt enable
// input capture 1 interupt enable
#  define USED_TIMSK_INTERRUPTS	(_BV(OCIE1A) | _BV(OCIE1B) | _BV(ICIE1))
#else
// output compare 1B is only needed for SWITCH
#  define USED_TIMSK_INTERRUPTS	(_BV(OCIE1A) | _BV(ICIE1))
#endif

// ----------------------------------------------------------------------
// IR receiver definitions
// ----------------------------------------------------------------------

#define	TIMEOUT		10500	// IR transmission timeout in us
#define	IR_MAX		35	// maximum number of IR data bytes
#define	IR_WRAP		16	// wrap area at end (for 67 byte NEC protocol)
#define	SCALING		10	// 9: 42.67 us period, 10: 85.33 us period
#define	LED		5	// PORTD bit number for LED
#define	PULLUP		3	// PORTD bit number for D- pullup

enum
{
	// Generic requests
	USBTINY_ECHO,		// echo test
	// IgorPlug-USB requests
	IGORPLUG_CLEAR,		// clear IR data
	IGORPLUG_READ,		// read IR data (wValue: offset)
	// LCD requests
	LCD_INSTR = 20,		// write instructions to LCD (via OUT)
	LCD_DATA,		// write data to LCD (via OUT)
	// SWITCH requests
	SWITCH_PROG = 30,	// program switch intervals
};

static	volatile byte_t	inpos = 0xff;	// read position for usb_in(), or 0xff
static	volatile byte_t	edges;		// incremented for each edge
static	struct			// IgorPlug-USB compatible data layout
{
	byte_t	length;		// length of data[]
	byte_t	count;		// incremented for each IR packet
	byte_t	offset;		// not used
	byte_t	data[IR_MAX];	// mark/space periods
}		ir;

// ----------------------------------------------------------------------
// LCD definitions
// ----------------------------------------------------------------------

#define LCD_LINES	2	// 1 or 2 lines

#define LCD_PORTNAME	B
#define D4		0
#define RS		4
#define RW		5
#define E		6

#define	LCD_DDR		_MERGE(DDR,  LCD_PORTNAME)
#define	LCD_PORT	_MERGE(PORT, LCD_PORTNAME)
#define	LCD_PIN		_MERGE(PIN,  LCD_PORTNAME)

#define	MASK_D4		(0xf << D4)
#define	MASK_RS		(1   << RS)
#define	MASK_RW		(1   << RW)
#define	MASK_E		(1   << E)

#ifndef	ICIE1
#define	ICIE1		TICIE1
#endif

// ----------------------------------------------------------------------
// IR detect / switch definitions
// ----------------------------------------------------------------------

#define	SWITCH_TIMEOUT	250000	// switch is activated that long per detect

#define SWITCH_PORTNAME	B
#define SWITCH_BIT	7

#define SWITCH_DDR	_MERGE(DDR,  SWITCH_PORTNAME)
#define SWITCH_PORT	_MERGE(PORT, SWITCH_PORTNAME)
#define SWITCH_PIN	_MERGE(PIN,  SWITCH_PORTNAME)

byte_t intervals_hi[IR_MAX] EEMEM;
byte_t intervals_lo[IR_MAX] EEMEM;

// The following macros are much shorter than the representive functions
// eeprom_{read,write}_byte.

// Read an eeprom entry. Address is x, result returned in register y
// - This is for ATtiny2313 - thus 8bit addresses
// - The check for write finish can be disabled to save 8 more bytes - if there
//   is a write pending, results would be undefined (which doesn't matter)
#define EEPROM_READ_8bit(x,y) do {		\
		while (EECR & _BV(EEPE))  ;	\
		EEAR=(x);			\
		EECR |= _BV(EERE);		\
		(y)=EEDR;			\
	} while (0)

// Write an eeprom entry. Address is x, value y
// - This is for ATtiny2313 - thus 8bit addresses
#define EEPROM_WRITE_8bit(x,y) do {		\
		while (EECR & _BV(EEPE))  ;	\
		cli();				\
		EEAR=(x);			\
		EEDR=(y);			\
		EECR |= _BV(EEMPE);		\
		EECR |= _BV(EEPE);		\
		sei();				\
	} while (0)

// ----------------------------------------------------------------------
// Handler for timer1 input capture interrupt: edge on IR input
// ----------------------------------------------------------------------
__attribute__((signal, NAKED))			// interrupts are DISABLED
extern	void    SIG_INPUT_CAPTURE1 ( void );
extern	void	SIG_INPUT_CAPTURE1 ( void )
{
	static	uint_t	prev;
	register uint_t stamp asm ("r20");	// not nearly enough to really use r20
	byte_t		delta;
	byte_t		e;

	NAKED_CODE (asm volatile(
		"push	r23\n"
		"in	r23,__SREG__\n"
		"push	r20\n"
		"push	r21\n"
	:: ));
	TIMSK = 0;				// disable both IR interrupts
	stamp = ICR1;				// get time stamp
	sei();					// allow USB interrupt
	// clobber all function-usable registers but the ones already saved.
	// this will trigger the register allocator to use those saved registers
	// for stamp.
	NAKED_CODE (asm volatile ("" : : :
		"r18","r19","r22","r23","r24","r25","r26","r27","r30","r31"
	));
	NAKED_CODE (asm volatile(
		"push	r24\n"
		"push	r25\n"
		"push	r18\n"
		"push	r19\n"
		"push	r30\n"
		"push	r31\n"
	:: ));
	TCCR1B ^= _BV(ICES1);			// toggle edge detector
	PORTD |= _BV(LED);			// switch LED on
	delta = (stamp - prev + (1 << (SCALING-CLK_DIV_SHIFT-1))) >> (SCALING-CLK_DIV_SHIFT);
	asm volatile("" : : "r"(delta) );	// calculate delta (gcc-4.1.0)
	prev  = stamp;
#if	US_2_CLK(TIMEOUT) > 65535
#	error TIMEOUT too big
#endif
	OCR1A = stamp + US_2_CLK(TIMEOUT);	// update timeout value
	e = edges;
	if	( e > IR_MAX ) {
		e -= IR_WRAP;
	}
	cli();					// enter critical region
	if	( edges != 0xff )		// packet should not be ignored?
	{
		if	( inpos == 0xff )	// update ir only when USB idle
		{
			ir.length = 0;		// discard previous packet
			if	( e > 0 )
			{
				ir.data[e - 1] = delta;
			}
		}
		edges = e + 1;
	}
	sei();					// allow USB interrupt
	NAKED_CODE (asm volatile(
		"pop	r31\n"
		"pop	r30\n"
		"pop	r19\n"
		"pop	r18\n"
	:: ));
	cli();
	TIMSK = USED_TIMSK_INTERRUPTS;		// reenable IR interrupts
	NAKED_CODE (asm volatile(
		"pop	r25\n"
		"pop	r24\n"
		"pop	r21\n"
		"pop	r20\n"
		"out	__SREG__,r23\n"
		"pop	r23\n"
		"reti\n"
	:: ));
}

// ----------------------------------------------------------------------
// Handler for timer1 output compare A interrupt: IR transmission timeout
// ----------------------------------------------------------------------
// naked interrupt handler in principle only necessary if SWITCH_PRESENT
__attribute__((signal, NAKED))			// interrupts are DISABLED
extern	void	SIG_OUTPUT_COMPARE1A ( void );
extern	void	SIG_OUTPUT_COMPARE1A ( void )
{
	byte_t len;
	NAKED_CODE (asm volatile(
		"push	r23\n"
		"in	r23,__SREG__\n"
	:: ));
	TIMSK = 0;				// disable both IR interrupts
	sei();					// allow USB interrupt
	NAKED_CODE (asm volatile(
		"push	r18\n"
		"push	r19\n"
		"push	r24\n"
		"push	r25\n"
		"push	r30\n"
		"push	r31\n"
	:: ));
	PORTD &= ~_BV(LED);			// switch LED off
	cli();					// enter critical region
	len = edges;
	edges = 0;
	if	(  len >= 4			// at least two pulses (NEC)
		&& len != 0xff			// packet should not be ignored
		&& inpos == 0xff		// update ir only when USB idle
		)
	{
		ir.count++;
		len--;
		ir.length = len;		// new packet is complete
#if	SWITCH_PRESENT
		sei();				// allow USB interrupt
		// This part is NOT racy. IR interrupts are still off.
		byte_t i = 0;
		for (i = 0; i < len; i++) {
			byte_t v, e;
			v = ir.data[i];
			//e = eeprom_read_byte (&intervals_hi[i]);
			EEPROM_READ_8bit(i,e);
			if (v > e)
				goto notfound;
			//e = eeprom_read_byte (&intervals_lo[i]);
			EEPROM_READ_8bit(i+IR_MAX,e);
			if (e == 0)
				break;
			if (v < e)
				goto notfound;
		}
#if	US_2_CLK(SWITCH_TIMEOUT) > 65535
#	error SWITCH_TIMEOUT too big
#endif
		OCR1B = ICR1 + US_2_CLK(SWITCH_TIMEOUT);
		SWITCH_PORT |= _BV(SWITCH_BIT);	// switch on
notfound:	(void)0;
#endif
	}
	sei();					// allow USB interrupt
	TCCR1B &= ~_BV(ICES1);			// reset to negative edge
	NAKED_CODE (asm volatile(
		"pop	r31\n"
		"pop	r30\n"
	:: ));
	cli();
	TIMSK = USED_TIMSK_INTERRUPTS;		// reenable IR interrupts
	NAKED_CODE (asm volatile(
		"pop	r25\n"
		"pop	r24\n"
		"pop	r19\n"
		"pop	r18\n"
		"out	__SREG__,r23\n"
		"pop	r23\n"
		"reti\n"
	:: ));
}

#if	SWITCH_PRESENT
// ----------------------------------------------------------------------
// Nuke switch after 250ms
// ----------------------------------------------------------------------
__attribute__((signal, NAKED))			// interrupts are DISABLED
extern	void	SIG_OUTPUT_COMPARE1B ( void );
extern	void	SIG_OUTPUT_COMPARE1B ( void )
{
	SWITCH_PORT &= ~_BV(SWITCH_BIT);	// switch off
	NAKED_CODE (asm volatile ("reti\n" :: ));
}

// ----------------------------------------------------------------------
// Initialize the Switch.
// ----------------------------------------------------------------------
static	void	switch_init ( void )
{
	SWITCH_DDR  |= _BV(SWITCH_BIT);
	SWITCH_PORT &= ~_BV(SWITCH_BIT);
}
#endif

// ----------------------------------------------------------------------
// Initialize the IR receiver.
// ----------------------------------------------------------------------
static	void	ir_init ( void )
{
	DDRD |= _BV(LED);
	TCCR1B = 
#if	SCALING-CLK_DIV_SHIFT >= 3
		 _BV(ICNC1)	// noise canceler, trigger on negative edge
#else
#	warn "Not enough cycles for input capture noise canceler"
		 0
#endif
#if	CLK_DIV_SHIFT == 3
	       | _BV(CS11);		// clock source clk/8
#elif	CLK_DIV_SHIFT == 6
	       | _BV(CS11) | _BV(CS10);	// clock source clk/64
#else
#	error "Bad CLK_DIV_SHIFT"
#endif
	TIMSK  = USED_TIMSK_INTERRUPTS;	// enable interrupts
}

#if	LCD_PRESENT
// ----------------------------------------------------------------------
// Delay <count> times 100 microseconds.
// ----------------------------------------------------------------------
static	void	lcd_delay100u ( byte_t count )
{
	asm volatile(
		"0:	ldi	r25, 240\n"
		"1:	rjmp	2f\n"
		"2:	dec	r25\n"
		"	brne	1b\n"	// 240 * 5 cycles = 100us
		"	dec	r24\n"
		"	brne	0b\n"
	);
}

// ----------------------------------------------------------------------
// Write 4 bits to the LCD.
// ----------------------------------------------------------------------
static	void	lcd_write4 ( byte_t data )
{
	LCD_PORT  = (LCD_PORT & ~ MASK_D4) | (data << D4);
	LCD_PORT |= MASK_E;		// E high period: > 230ns
	LCD_DDR   = MASK_RS | MASK_RW | MASK_E | MASK_D4; // Dx setup time: > 80ns
	LCD_PORT &= ~ MASK_E;		// hold time: > 10ns
	LCD_DDR   = MASK_RS | MASK_RW | MASK_E;
}

// ----------------------------------------------------------------------
// Read 4 bits from the LCD.
// ----------------------------------------------------------------------
static	byte_t	lcd_read4 ( void )
{
	byte_t	save;
	byte_t	data;

	save = LCD_PORT;
	LCD_PORT = MASK_RW;		// address setup time: > 40ns
	LCD_PORT |= MASK_E;		// E high period: > 230ns
	asm volatile("rjmp 0f\n0:");
	data = (LCD_PIN & MASK_D4) >> D4;// data access time: > 160ns
	LCD_PORT = save;
	return data;
}

// ----------------------------------------------------------------------
// Write a byte to the LCD.
// ----------------------------------------------------------------------
static	void	lcd_write ( byte_t b )
{
	byte_t	h;

	do
	{
		h = lcd_read4();
		(void) lcd_read4();
	} while	( h & 0x08 );
	lcd_write4( b >> 4 );
	lcd_write4( b & 15 );
}

// ----------------------------------------------------------------------
// Initialize the LCD.
// ----------------------------------------------------------------------
static	void	lcd_init ( void )
{
	LCD_DDR = MASK_RS | MASK_RW | MASK_E;
	LCD_PORT = 0;
	//lcd_delay100u( 150 );	-- skipped because of RESET delay
	lcd_write4( 3 );
	lcd_delay100u( 41 );
	lcd_write4( 3 );
	lcd_delay100u( 1 );
#if	LCD_INIT_PRESENT
	lcd_write( 0x32 );	// switch to 4-bit mode
	lcd_write( 0x20 +	// function set: 4-bit mode + #lines
		   (LCD_LINES-1)*0x08 );
	lcd_write( 0x01 );	// clear display
	lcd_write( 0x06 );	// entry mode: increment, no shift
	lcd_write( 0x0c );	// display control: on, no cursor
#endif
}
#endif	/* LCD_PRESENT */

// ----------------------------------------------------------------------
// Handle a non-standard SETUP packet.
// ----------------------------------------------------------------------
extern	byte_t	usb_setup ( byte_t data[8] )
{
	byte_t	req;
	byte_t	r;

	r = 0;

	// Generic requests
	req = data[1];
	if	( req == USBTINY_ECHO )
	{
		r = 8;
	}

	// IgorPlug-USB requests
	if	( req == IGORPLUG_CLEAR )
	{
		ir.length = 0;
		inpos = 0xff;
	}
	if	( req == IGORPLUG_READ )
	{
		cli();
		if	( ir.length > 0 )
		{
			inpos = data[4];
			r = 0xff;	// call usb_in() to get the data
		}
		else
		{
			data[0] = 0;
			r = 1;
		}
		sei();
	}

#if	LCD_PRESENT
	// LCD requests
	if	( req == LCD_INSTR )
	{
		LCD_PORT &= ~ MASK_RS;	// data will be received by usb_out()
	}
	if	( req == LCD_DATA )
	{
		LCD_PORT |= MASK_RS;	// data will be received by usb_out()
	}
#endif
#if	SWITCH_PROG_PRESENT && LCD_PRESENT
	if	( req == SWITCH_PROG )
	{
		inpos = 0;		// data will be received by usb_out()
	}
#endif

	return r;
}

// ----------------------------------------------------------------------
// Handle an IN packet.
// ----------------------------------------------------------------------
extern	byte_t	usb_in ( byte_t* data, byte_t len )
{
	byte_t	n;
	byte_t	max;

	max = ir.length + 3;
	n = 0;
	while	( n < len )
	{
		if	( inpos >= max )	// end of packet?
		{
			cli();
			if	( edges )	// next packet already started?
			{
				edges = 0xff;	// ignore remainder of packet
			}
			inpos = 0xff;		// reenable receiver
			sei();
			break;
		}
		data[n++] = (& ir.length)[inpos++];
	}
	return n;
}

// ----------------------------------------------------------------------
// Handle an OUT packet.
// ----------------------------------------------------------------------
extern	void	usb_out ( byte_t* data, byte_t len )
{
#if	SWITCH_PROG_PRESENT
#	if	LCD_PRESENT
	if (! inpos) {
#	endif
		// Protocol: byte 0: pos  byte 1: data
		EEPROM_WRITE_8bit (data[0], data[1]);
		return;
#	if	LCD_PRESENT
	}
#	endif
#endif
#if	LCD_PRESENT
	while	( len )
	{
		lcd_write( *data++ );
		len--;
	}
#endif
}

// ----------------------------------------------------------------------
// Main
// ----------------------------------------------------------------------
__attribute__((naked))		// suppress redundant SP initialization
extern	int	main ( void )
{
	PORTD |= _BV(PULLUP);
	DDRD  |= _BV(PULLUP);	// enable pullup on D-
#if	LCD_PRESENT
	lcd_init();
#endif
#if	SWITCH_PRESENT
	switch_init();
#endif
	usb_init();
	ir_init();
	for	( ;; )
	{
		usb_poll();
	}
	return 0;
}
